#!/usr/bin/perl
require 5.001;
use strict;
use Date::Manip qw(ParseDate UnixDate);

require './cgi-lib.pl';


###################################################################
# Routines for safely reading input values
# TODO: Add robust date and time handling
###################################################################
sub ReadMode(\%) {
  # check for a valid mode
  my $in = shift;     # in is a reference to a hash
  my $mode = $in->{"mode"};
  if ($mode =~ /^(edit|submit|summary|details)$/) {
    return $mode;
  } 
  return 'summary';
}

sub ReadText(\%$) {
  my $in = shift;     # in is a reference to a hash
  my $key = shift;
  my $val = $in->{$key};
  $val =~ s/\n|\r/ /gm;
  $val =~ s/</&lt;/g;
  $val =~ s/>/&gt;/g;
  return $val;
}

sub ReadTextArea(\%$) {
  my $in = shift;     # in is a reference to a hash
  my $key = shift;
  my $val = $in->{$key};
  if (!defined($val)) {
    $val = "";
  }
  $val =~ s/</&lt;/gm;
  $val =~ s/>/&gt;/gm;
  return $val;
}

sub ReadTextArea(\%$) {
  my $in = shift;     # in is a reference to a hash
  my $key = shift;
  my $val = $in->{$key};
  if (!defined($val)) {
    $val = "";
  }
  $val =~ s/</&lt;/gm;
  $val =~ s/>/&gt;/gm;
  return $val;
}

sub ReadNumber(\%$) {
  my $in = shift;     # in is a reference to a hash
  my $key = shift;
  my $val = $in->{$key};
  if (!defined($val) || $val eq "") {
    return "";
  }
  if ($val =~ /^\d+$/) {
    return $val;
  }
  return "error"; 
}

sub ReadFFMinutes(\%) {
  my $in = shift;     # in is a reference to a hash
  my $val = $in->{'ffminutes'};
  # FIXME: An error check is missing here
  if ($val < 0 || $val > 5) {
    return "error";
  }
  return $val; 
}

sub ReadFFSeconds(\%) {
  my $in = shift;     # in is a reference to a hash
  my $val = $in->{'ffseconds'};
  if (!$val =~ /^\d+$/) {
    return "error";
  }
  if ($val < 0 || $val > 59) {
    return "error";
  }
  return $val; 
}

sub ReadId(\%) {
  my $in = shift;     # in is a reference to a hash
  my $id = $in->{'id'};
  if ( $id =~ /^\d+$/ ) {
    return $id;
  } 
  return '';
}

sub ReadDate(\%) {
  my $in = shift;     # in is a reference to a hash
  my $date = $in->{'date'};
  my $date = ParseDate($date);
  if ($date) {
    my ($y, $m, $d) = ($date =~ m/^(\d{4})(\d{2})(\d{2})/);
    return "$y-$m-$d";
  }
  return "";
}

sub ReadTime(\%) {
  my $in = shift;     # in is a reference to a hash
  my $time = $in->{'time'};
  my $time = ParseDate($time);
  if ($time) {
    my ($h, $m, $s) = ($time =~ m/(\d{2}):(\d{2}):(\d{2})$/);
    return "$h:$m";
  }
  return "";
}

sub ShowFileWithSubstitutions($\%) {
  my $filename = shift;
  my $subs = shift;
  open(my $fp, "<", $filename);
  while (my $line = <$fp>) {
    while (my ($var) = $line =~ /__(\w+)__/) {
      my $sub = $subs->{$var};
      if (!$sub) { $sub = ''; }
      $line =~ s/__($var)__/$sub/;
    }
    print($line);
  }
  close($fp);
}

sub ShowFile($) {
  my $filename = shift;
  my %dummy;
  ShowFileWithSubstitutions($filename, %dummy)
}

###################################################################
# Routines for database IO
###################################################################
sub SaveReading(\%) {
  my $reading = shift;
  open(my $fp, ">>", "bplog.txt");
  my $id = $reading->{'id'};
  for my $field (keys(%$reading)) {
    if ($field ne 'id') {
      my $value = $reading->{$field};
      $value =~ s/\n/<n>/gm;
      $value =~ s/\r/<r>/gm;
      print $fp ("$id:$field:$value\n");
    }
  }
  close($fp);
}

# read the blood pressure log into a hash (database)
# database is keyed by id
# each value in database is (a reference to) a hash
# that maps field names (date, time, diastolic, sysstolic) onto readings
sub ReadDatabase(\%) {
  my $database = shift;
  open(my $fp, "<", "bplog.txt");
  while (my $line = <$fp>) {
    if (my ($id, $field, $value) 
          = ($line =~ /^(\w+):(\w+):(.*)$/) ) {
      my $reading = $database->{$id};
      if (!$reading) {  # add a new key to %database
        $reading = {'id' => $id};
        $database->{$id} = $reading;
      }
      $value =~ s/<r>/\r/gm;
      $value =~ s/<n>/\n/gm;
      $reading->{$field} = $value;
    }
  }
  close($fp);
}

# debugging routine
sub DumpDatabase(\%) {
  my $database = shift;
  for my $id (keys(%$database)) {
    print("Reading $id\n");
    my $reading = $database->{"$id"};
    for my $field (keys(%$reading)) {
      my $value = $reading->{$field};
      print("  $field = $value\n");
    }
  }
  return 0;
}


# Return a unique string
sub GenerateNewId() {
  return (int(rand(1000000000)));
}

sub ShowSummary() {
  # summary mode shows the all measurements to date
  my %database;
  ReadDatabase(%database);
  ShowFile('summary-head.html');
  # sort the database by date and time
  my @sorteddb 
    = sort { $a->{'date'}.$a->{'time'} cmp
             $b->{'date'}.$b->{'time'} } (values(%database));
  for my $reading (@sorteddb) {
    print("  <tr>\n");
    print("    <td>$reading->{'date'}</td>\n"); 
    print("    <td>$reading->{'time'}</td>\n"); 
    print("    <td>$reading->{'dropzone'}</td>\n"); 
    print("    <td>$reading->{'exitalt'}</td>\n"); 
    print("    <td><a href=\"?mode=edit&amp;id=$reading->{'id'}\">edit</a></td>\n"); 
    print("    <td><a href=\"?mode=details&amp;id=$reading->{'id'}\">details</a></td>\n"); 
    print("  </tr>\n");
  }
  ShowFile('summary-tail.html');
}

MAIN: {
  my %in;
  ReadParse(\%in);
  print("Content-type: text/html\n\n");

  my $mode = ReadMode(%in);
  if ($mode eq 'summary') {
    ShowSummary();
  } elsif ($mode eq 'details') {
    # Show the details of a specific jump
    my $id = ReadId(%in); 
    if (!$id) {
      ShowFile("error.html");
      exit(0);
    }
    ReadDatabase(my %database);
    my $reading = $database{$id};  
    if (!$reading) {
      ShowFile("error.html");
      exit(0);
    }
    # find next and previous dives
    my @sorteddb 
      = sort { $a->{'date'}.$a->{'time'} cmp
               $b->{'date'}.$b->{'time'} } (values(%database));
    my $idx = 0;
    for (my $i = 0; $i < scalar(@sorteddb); $i++) {
      if ($sorteddb[$i]->{'id'} eq $id)  {
        $idx = $i;  $i = scalar(@sorteddb); 
      }
    }
    my $prev = "<< Prev";
    if ($idx > 0) {
      my $previd = $sorteddb[$idx-1]->{'id'};
      $prev = "<a href=\"?mode=details&amp;id=$previd\">$prev</a> "; 
    }
    my $next = "Next >>";
    if ($idx + 1 < scalar(@sorteddb)) {
      my $nextid = $sorteddb[$idx+1]->{'id'};
      $next = "<a href=\"?mode=details&amp;id=$nextid\">$next</a> "; 
    }
    $reading->{'ro'} = "readonly";
    $reading->{'footer'} = 
      $prev .  " <a href=\"?mode=edit&amp;id=$id\">Edit</a> " .  $next .
      ' <a href="?mode=summary">Summary</a>';
    ShowFileWithSubstitutions("edit.html", %$reading);
  } elsif ($mode eq 'edit') {
    # edit mode allows editing of a jump
    my $id = ReadId(%in); 
    my $reading = { };
    if ($id) { # editing an existing jump
      ReadDatabase(my %database);
      $reading = $database{$id};  
      if (!$reading) {
        ShowFile("error.html");
        exit(0);
      }
    } else {   # adding a new jump
      $id = GenerateNewId();
      $reading->{'id'} = $id;
    }
    $reading->{'footer'} =
     '<input type="submit" value="Save Jump Data">'.
     '<a href="?mode=summary">Cancel</a>';
    ## ffminutes and ffseconds are kind of annoying
    my $ffminutes = "<select id='ffminutes' name='ffminutes'>";
    for (my $i = 0; $i < 6; $i++) {
      if ($reading->{'ffminutes'} == $i) {
        $ffminutes .= "<option value='$i' selected>$i</option>";
      } else {
        $ffminutes .= "<option value='$i'>$i</option>";
      }
    }
    $ffminutes .= "</select>";
    $reading->{'ffminutes'} = $ffminutes;
    my $ffseconds = "<select id='ffseconds' name='ffseconds'>";
    for (my $i = 0; $i < 60; $i++) {
      if ($reading->{'ffseconds'} == $i) {
        $ffseconds .= "<option value='$i' selected>$i</option>";
      } else {
        $ffseconds .= "<option value='$i'>$i</option>";
      }
    }
    $ffseconds .= "</select>";
    $reading->{'ffseconds'} = $ffseconds;

    ShowFileWithSubstitutions("edit.html", %$reading);
  } elsif ($mode eq 'submit') {
    # submit mode occurs when an entry is submitted
    my %reading;
    my $error = "";
    $reading{'id'} = ReadId(%in);
    if (!$reading{'id'}) {
      ShowFile("error.html");
      exit(0);
    }
    $reading{'date'} = ReadDate(%in);
    if (!$reading{'date'}) {
      $error = $error."<div class=\"error\">".
	"Invalid date, not in format (YYYY-MM-DD)</div>\n ";
    }
    $reading{'time'} = ReadTime(%in);
    if (!$reading{'time'}) {
       $error = $error."<div class=\"error\">".
	"Invalid time, not in format (HH:MM)</div>\n ";
    }
    $reading{'dropzone'} = ReadText(%in, 'dropzone');
    if (!$reading{'dropzone'}) {
       $error = $error."<div class=\"error\">".
	"The dropzone field is mandatory";
    }
    $reading{'aircraft'} = ReadText(%in, 'aircraft');
    $reading{'exitalt'} = ReadNumber(%in, 'exitalt');
    if ($reading{'exitalt'} eq 'error') {
        $error = $error."<div class=\"error\">".
	"Invalid exit altitude, not a number</div>\n ";
    }
    $reading{'deployalt'} = ReadNumber(%in, 'deployalt');
    if ($reading{'deployalt'} eq 'error') {
        $error = $error."<div class=\"error\">".
	"Invalid deploy altitude, not a number</div>\n ";
    }
    $reading{'ffminutes'} = ReadFFMinutes(%in);
    if ($reading{'ffminutes'} eq 'error') {
        $error = $error."<div class=\"error\">".
	"Invalid freefall time (minutes)</div>\n ";
    }
    $reading{'ffseconds'} = ReadFFSeconds(%in);
    if ($reading{'ffseconds' eq 'error'}) {
        $error = $error."<div class=\"error\">".
	"Invalid freefall time (seconds)</div>\n ";
    }
    $reading{'manoeuvres'} = ReadText(%in, 'manoeuvres');
    $reading{'otherjumpers'} = ReadText(%in, 'otherjumpers');
    $reading{'comments'} = ReadTextArea(%in, 'comments');
    $reading{'signature'} = ReadText(%in, 'signature');
    if ($error) {
      $reading{'error'} = $error; 
      $reading{'footer'} =
        '<input type="submit" value="Save Jump Data">'.
        '<a href="?mode=summary">Cancel</a>';
      ShowFileWithSubstitutions("edit.html", %reading);
    } else {
      SaveReading(%reading);
      ShowSummary();
    }
  } else {
    ShowFile("error.html");
  }
} 
